from .issue_provider import AzureIssueProvider
from .story_point_extractor import AzureStoryPointExtractor
from .worklog_extractor import (
    AzureStatusChangeWorklogExtractor,
    AzureIssueTotalSpentTimeExtractor,
)



import base64
from abc import abstractmethod, ABC
from typing import Optional


class IssueProvider(ABC):

    @abstractmethod
    def get_issues(self) -> list:
        pass


class ProxyIssueProvider(IssueProvider):

    def __init__(self, issues: list) -> None:
        self.issues = issues

    def get_issues(self) -> list:
        return self.issues


class CachingIssueProvider(IssueProvider):

    def __init__(self, provider: IssueProvider, cache: Optional[dict] = None) -> None:
        self.cache = cache

        self.provider = provider
        self.query = getattr(provider, 'query', None)
        self.additional_fields = getattr(provider, 'additional_fields', None)


    def get_issues(self):
        cached_issues = self._search_in_cache()
        if cached_issues is not None:
            return cached_issues

        issues = self.provider.get_issues()
        self._set_in_cache(issues)
        return issues

    def _search_in_cache(self):
        if self.cache is None:
            return None

        # if query with same additional_fields is already present in cache
        cache_key_with_fields = self._create_cache_key_for_query()
        if self._is_key_in_cache(cache_key_with_fields):
            return self._get_from_cache(cache_key_with_fields)

        # searching for cached responses with not less additional_fields (superset)
        partial_key = self._create_partial_cache_key()
        for key in self._get_all_cache_keys():
            if key.startswith(partial_key):
                if self.additional_fields is None:
                    return self._get_from_cache(key)
                else:
                    all_fields_present = True
                    for field in self.additional_fields:
                        if "_" + field not in key:
                            all_fields_present = False
                    if all_fields_present:
                        return self._get_from_cache(key)

        return None

    def _get_from_cache(self, cache_key):
        return self.cache.get(cache_key)

    def _set_in_cache(self, issues):
        if self.cache is None:
            return
        self.cache[self._create_cache_key_for_query()] = issues

    def _is_key_in_cache(self, cache_key):
        return cache_key in self.cache

    def _get_all_cache_keys(self):
        return self.cache.keys()

    def _create_cache_key_for_query(self):
        if self.additional_fields is None:
            return self._create_partial_cache_key()
        return self._create_partial_cache_key() + "_".join(self.additional_fields)

    def _create_partial_cache_key(self):
        if self.query is None:
            return "none_query||"
        return base64.b64encode(self.query.encode("ascii")).decode("ascii") + "||"


import math
from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED
from typing import Iterable, List, Optional

from azure.devops.v7_1.work_item_tracking.models import Wiql

from data_providers.issue_provider import IssueProvider


class AzureIssueProvider(IssueProvider):
    DEFAULT_FIELDS = [
        'System.Title',
        'System.WorkItemType',
        'System.State',
        'System.CreatedDate',
        'System.AssignedTo',

        'Microsoft.VSTS.Scheduling.StoryPoints',
        'Microsoft.VSTS.Common.ClosedDate',
    ]

    def __init__(self, azure_client, query: str, additional_fields: Optional[Iterable[str]] = None,
                 page_size: int = 200, thread_pool_executor: Optional[ThreadPoolExecutor] = None) -> None:
        self.azure_client = azure_client
        self.query = query.strip()
        self.additional_fields = list(additional_fields) if additional_fields is not None else list(self.DEFAULT_FIELDS)
        self.page_size = max(1, page_size)
        self.thread_pool_executor = thread_pool_executor

    def get_issues(self) -> list:
        query = Wiql(query=self.query)
        query_result = self.azure_client.query_by_wiql(query)

        work_item_ids = [ref.id for ref in (query_result.work_items or [])]

        fetched_issues = []
        if not work_item_ids:
            return fetched_issues

        total_ids = len(work_item_ids)
        total_batches = math.ceil(total_ids / float(self.page_size))

        if self.thread_pool_executor is None:
            fetched_issues = self._fetch_issue_sync(work_item_ids, total_batches, total_ids)
        else:
            fetched_issues = self._fetch_issue_concurrently(work_item_ids, total_batches, total_ids)

        self._attach_changelog_history(fetched_issues)
        return fetched_issues

    def _fetch_issue_sync(self, work_item_ids: List[int], total_batches: int, total_ids: int) -> List:
        issues = []
        for batch_index in range(total_batches):
            batch_start = batch_index * self.page_size
            batch_end = min(batch_start + self.page_size, total_ids)
            batch_ids = work_item_ids[batch_start:batch_end]
            wis = self.azure_client.get_work_items(ids=batch_ids, fields=self.additional_fields)
            issues.extend(wis or [])
        return issues

    def _fetch_issue_concurrently(self, work_item_ids: List[int], total_batches: int, total_ids: int) -> List:
        issues = []
        futures = []
        for batch_index in range(total_batches):
            batch_start = batch_index * self.page_size
            batch_end = min(batch_start + self.page_size, total_ids)
            batch_ids = work_item_ids[batch_start:batch_end]
            futures.append(
                self.thread_pool_executor.submit(self.azure_client.get_work_items, ids=batch_ids, fields=self.additional_fields))
        done = wait(futures, return_when=ALL_COMPLETED).done
        for done_feature in done:
            issues.extend(done_feature.result() or [])
        return issues

    def _attach_changelog_history(self, issues: List[object]):

        def fetch_changelog_history(issue):
            issue.fields['CustomExpand.WorkItemUpdate'] = self.azure_client.get_updates(issue.id)

        if self.thread_pool_executor is None:
            for issue in issues:
                fetch_changelog_history(issue)
        else:
            futures = [self.thread_pool_executor.submit(fetch_changelog_history, issue) for issue in issues]
            wait(futures, return_when=ALL_COMPLETED)



from data_providers.utils.enums import VelocityTimeUnit, HealthStatus, SeniorityLevel
from data_providers.utils.query_utils import JiraIssueSearchQueryBuilder, TimeRangeGenerator
from data_providers.utils.story_point_utils import TShirtMapping



from data_providers.issue_provider import IssueProvider, CachingIssueProvider, ProxyIssueProvider
from data_providers.story_point_extractor import StoryPointExtractor
from data_providers.worklog_extractor import WorklogExtractor, ChainedWorklogExtractor, IssueTotalSpentTimeExtractor
